/*----------------------------------------------------------------------------
    TAPESTREA: Techniques And Paradigms for Expressive Synthesis, 
               Transformation, and Rendering of Environmental Audio
      Engine and User Interface

    Copyright (c) 2006 Ananya Misra, Perry R. Cook, and Ge Wang.
      http://taps.cs.princeton.edu/
      http://soundlab.cs.princeton.edu/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// name: ui_audio.cpp
// desc: birdbrain ui audio
//
// authors: Ananya Misra (amisra@cs.princeton.edu)
//          Ge Wang (gewang@cs.princeton.edu)
//          Perry R. Cook (prc@cs.princeton.edu)
// date: Winter 2004
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <memory.h>
#include "ui_audio.h"
#include "taps_treesynth.h"
using namespace std;


// static variables
AudioCentral * AudioCentral::our_instance = NULL;


//-----------------------------------------------------------------------------
// name: cb()
// desc: callback
//-----------------------------------------------------------------------------
static int cb( char * buffer, int buffer_size, void * user_data )
{
    AudioCentral * ac = (AudioCentral *)user_data;
    SAMPLE * buf = ac->m_buffer;
    SAMPLE * buf2 = ac->m_buffer2;
    t_TAPUINT num_frames = ac->m_num_frames;
    t_TAPUINT num_samples = num_frames * ac->m_channels;
    t_TAPBOOL yes = FALSE;

    // zero out the buffer 
    assert( num_frames == buffer_size );
    memset( buffer, 0, sizeof(SAMPLE) * num_samples );

    // lock
    ac->m_mutex.acquire();
    // get samples
    if( ac->m_channels == 2 )
		yes = ac->m_bus->stick( (SAMPLE *)buffer, num_frames );
    else
		yes = ac->m_bus->multitick( buf2, (SAMPLE *)buffer, num_frames );

	// release
    ac->m_mutex.release();
    
    return 0;
}




//-----------------------------------------------------------------------------
// name: mono2stereo()
// desc: on same buffer
//-----------------------------------------------------------------------------
void AudioSrc::mono2stereo( SAMPLE * self, t_TAPUINT num_frames )
{
    // start from end of mono
    SAMPLE * src = self + num_frames - 1;
    SAMPLE * dest = self + (num_frames*2) - 2;

    // copy
    while( src >= self )
    {
        *dest = *(dest+1) = *src;
        src--; dest -= 2;
    }
}


//-----------------------------------------------------------------------------
// name: stereo2mono()
// desc: on same buffer
//-----------------------------------------------------------------------------
void AudioSrc::stereo2mono( SAMPLE * self, t_TAPUINT num_frames )
{
    // start from end of mono
    SAMPLE * src = self;
    SAMPLE * end = self + (num_frames*2);
    SAMPLE * dest = self;

    // copy
    while( src < end )
    {
        *dest = (*src + *(src+1))/2.0f;
        src += 2; dest++;
    }

    // zero
    while( dest < end )
        *dest++ = 0.0f;
}


//-----------------------------------------------------------------------------
// name: mono2stereo()
// desc: ...
//-----------------------------------------------------------------------------
void AudioSrc::mono2stereo( const SAMPLE * mono, t_TAPUINT num_frames, SAMPLE * stereo )
{
    SAMPLE * p = stereo, * end = stereo + (num_frames*2);
    // loop
    while( p < end )
    {
        // copy left and right from mono
        *p = *(p+1) = *mono;
        // advance to next frame
        mono++; p += 2;
    }
}




//-----------------------------------------------------------------------------
// name: stereo2mono()
// desc: ...
//-----------------------------------------------------------------------------
void AudioSrc::stereo2mono( const SAMPLE * stereo, t_TAPUINT num_frames, SAMPLE * mono )
{
    const SAMPLE * s = stereo, * end = stereo + (num_frames*2);
    SAMPLE * p = mono;
    // loop
    while( s < end )
    {
        // average and copy
        *p = (*s + *(s+1)) / 2.0f;
        // advance to next frame
        p++; s += 2;
    }
}


//-----------------------------------------------------------------------------
// name: mtick
// desc: mono tick
//-----------------------------------------------------------------------------
t_TAPBOOL AudioSrc::mtick( SAMPLE * buffer, t_TAPUINT num_frames )
{
    SAMPLE * sbuffer = new SAMPLE[num_frames * 2];
    t_TAPBOOL ret = this->stick( sbuffer, num_frames );
    stereo2mono( sbuffer, num_frames, buffer );
    return ret;
}



//-----------------------------------------------------------------------------
// name: pan_buf()
// desc: you know, the mythical creature
//-----------------------------------------------------------------------------
void AudioSrc::pan_buf( SAMPLE * stereo, t_TAPUINT num_frames, t_TAPSINGLE where )
{
    t_TAPSINGLE left_gain = where <= .5f ? 1.0f : (1.0f - where) * 2;
    t_TAPSINGLE right_gain = where >= .5f ? 1.0f : where * 2;
    SAMPLE * s = stereo, * end = stereo + (num_frames*2);
    // left
    if( left_gain != 1.0f )
    {
        while( s < end )
        {
            *s *= left_gain;
            s += 2;
        }
    }

    // right
    if( right_gain != 1.0f )
    {
        // reset pointer
        s = stereo + 1;
        while( s < end )
        {
            *s *= right_gain;
            s += 2;
        }
    }
}



    
//-----------------------------------------------------------------------------
// name: gain_buf()
// desc: ...
//-----------------------------------------------------------------------------
void AudioSrc::gain_buf( SAMPLE * buffer, t_TAPUINT num_samples, SAMPLE gain )
{
    if( gain != 1.0f )
        for( t_TAPUINT i = 0; i < num_samples; i++ )
            buffer[i] *= gain;
}




AudioCentral * AudioCentral::instance()
{
    if( !our_instance )
        our_instance = new AudioCentral;

    return our_instance;
}




//-----------------------------------------------------------------------------
// name: init()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioCentral::init( t_TAPUINT srate, t_TAPUINT num_frames,
                             t_TAPUINT sig_frames, t_TAPUINT bus, t_TAPUINT channels )
{
    // if already started...
    if( m_audio != NULL )
    {
        BB_log( BB_LOG_SYSTEM, "[ui_audio]: already started..." );
        return FALSE;
    }

    // frame
    if( num_frames < sig_frames )
    {
        BB_log( BB_LOG_SYSTEM, "[ui_audio]: num_frames < sig_framds - bad!" );
        return FALSE;
    }

	// chief minister
	if( m_cachemanager != NULL )
	{
		BB_log( BB_LOG_SYSTEM, "[ui_audio]: cache manager already exists!" ); 
		return FALSE;
	}

    // log
    BB_log( BB_LOG_SYSTEM, "initializing audio engine..." );
    // push log
    BB_pushlog();

    // set srate
    m_srate = srate;
    // set num_frames
    m_num_frames = num_frames;
    // set sig_frames
    m_sig_frames = sig_frames;
    // number of buffers
    t_TAPINT num_buffers = 4;
	// number of channels for real-time audio output
	m_channels = (channels >= 0 && channels <= bus) ? channels : 2;

    // log
    BB_log( BB_LOG_SYSTEM, "sample rate: %d", srate );
    BB_log( BB_LOG_SYSTEM, "buffer size: %d", num_frames );
    BB_log( BB_LOG_SYSTEM, "signal size: %d", sig_frames );
    BB_log( BB_LOG_SYSTEM, "num buffers: %d", num_buffers );
    BB_log( BB_LOG_SYSTEM, "channels: %d", m_channels );

    // open the device
    try
    {
        int size = m_num_frames;
        // open the audio device for capture and playback
        m_audio = new RtAudio( 0, m_channels, 0, 0, RTAUDIO_FLOAT32,
            m_srate, &size, num_buffers );
        m_num_frames = size;
    }
    catch( RtError & e )
    {
        // exception
        BB_log( BB_LOG_SYSTEM, "[ui_audio]: %s", e.getMessage().c_str() );
        BB_log( BB_LOG_SYSTEM, "[ui_audio]: cannot open audio device for playback" );
        // zero
        m_audio = NULL;

        return FALSE;
    }

    // allocate buffers
    m_buffer = new SAMPLE[m_num_frames * m_channels];
    m_buffer2 = new SAMPLE[m_num_frames * m_channels];

    // log
    BB_log( BB_LOG_SYSTEM, "allocating parallel bus..." );
    // draw parallel
    m_bus = new AudioBusParallel( m_num_frames, m_sig_frames );
    m_bus->m_delete = FALSE;
    
    // log
    BB_log( BB_LOG_SYSTEM, "allocating %d audio buses...", bus );
    // push log
    BB_pushlog();
//	fprintf( stdout, "buses:\n" );
	for( t_TAPUINT i = 0; i < bus; i++ )
    {
        // log
        BB_log( BB_LOG_INFO, "initializing audio bus %d...", i );
        AudioBus * audbus = new AudioBus( m_num_frames );
        audbus->m_delete = FALSE;
        m_bus->m_src.push_back( audbus );
//      fprintf( stdout, "%i - %x\n", i, m_bus->m_src[i] );
    }
    // pop log
    BB_poplog();

	// cache manager
	m_cachemanager = new CacheManager; 
	BB_log( BB_LOG_SYSTEM, "initializing cache manager..." );

    // set the audio callback
    m_audio->setStreamCallback( cb, this );

    // log
    BB_log( BB_LOG_SYSTEM, "starting real-time audio..." );
    // start the audio
    m_audio->startStream( );

    // pop log
    BB_poplog();

    return TRUE;
}




//-----------------------------------------------------------------------------
// name: free()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioCentral::free()
{
    // if not started
    if( m_audio == NULL )
        return FALSE;

    // clear the queue
    this->stop_all();
    // clear bus
    SAFE_DELETE( m_bus );

    // stop the audio
    m_audio->stopStream();
    
    // free the audio device
    SAFE_DELETE( m_audio );

    // free buffers
	SAFE_DELETE_ARRAY( m_buffer );
	SAFE_DELETE_ARRAY( m_buffer2 );

	// delete cache manager
	SAFE_DELETE( m_cachemanager );

    return TRUE;
}




//-----------------------------------------------------------------------------
// name: bus()
// desc: ...
//-----------------------------------------------------------------------------
AudioBus * AudioCentral::bus( t_TAPUINT index )
{
    if( index < 0 || index >= m_bus->m_src.size() )
        return NULL;

    return (AudioBus *)m_bus->m_src[index];
}




//-----------------------------------------------------------------------------
// name: stop()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioCentral::stop_all()
{
    // audio
    if( m_audio == NULL ) return FALSE;

    // synchronize
    m_mutex.acquire();
    // stop
    for( long i = 0; i < m_bus->m_src.size(); i++ )
        ((AudioBus *)m_bus->m_src[i])->stop();
    // release
    m_mutex.release();

    return TRUE;
}




//-----------------------------------------------------------------------------
// name: stick()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioSrcTest::stick( SAMPLE * buffer, t_TAPUINT num_frames )
{
//    assert( m_done == FALSE );

    // are we done?
    if( m_stop_asap )
        return FALSE;

    // 1 second
    if( m_t > BirdBrain::srate() )
    {
        m_done = TRUE;
        return FALSE;
    }

    // samples
    for( t_TAPUINT i = 0; i < num_frames; i++ )
    {
        buffer[i*2] = .25 * sin( m_t * m_freq * 2.0 * PIE / BirdBrain::srate() );
        buffer[i*2+1] = buffer[i*2];
        m_t += 1.0;
    }

    // gain
    gain_buf( buffer, num_frames * 2, m_gain );
    // pan
    pan_buf( buffer, num_frames, m_pan );

    return TRUE;
}




//-----------------------------------------------------------------------------
// name: AudioSrcFile()
// desc: ...
//-----------------------------------------------------------------------------
AudioSrcFile::AudioSrcFile()
    : AudioSrc()
{
    m_buffer = NULL;
    m_buffer_size = 0;
    sf = NULL;
    slide = NULL;
    slide_locally = NULL;
    signal_scale = 1.0f;
#ifdef SECRET_RABBIT_CODE
	sr = NULL;
#endif
	sr_gain = 1.0f;
}




//-----------------------------------------------------------------------------
// name: ~AudioSrcFile()
// desc: ...
//-----------------------------------------------------------------------------
AudioSrcFile::~AudioSrcFile()
{
    if( sf )
    {
        this->close();
    }

    if( on ) *on = false;
}




///-----------------------------------------------------------------------------
// name: open()
// desc: ...
//-----------------------------------------------------------------------------
void AudioSrcFile::rewind()
{
	AudioSrc::rewind();
	sf_seek( sf, 0, SEEK_SET );
#ifdef SECRET_RABBIT_CODE
	if( sr )
		src_reset( sr );
#endif
}




///-----------------------------------------------------------------------------
// name: open()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioSrcFile::open( const char * filename, t_TAPUINT start, t_TAPUINT len, 
							 t_TAPBOOL scale, int id, t_TAPBOOL usecache )
{
    // sanity
    if( sf )
    {
       BB_log( BB_LOG_WARNING, "[audiosrcfile]: file '%s' already open...", m_filename.c_str() );
       return FALSE;
    }

    // open the file
//    sf = sf_open( filename, SFM_READ, &sf_info );
	if( usecache && AudioCentral::instance()->m_cachemanager->isopen( filename ) > id )
	{
		// if open, just add to the count
		sf = AudioCentral::instance()->m_cachemanager->opensf( filename, SFM_READ, &sf_info, id );
	}
	else
	{
		sf = sf_open( filename, SFM_READ, &sf_info );
		if( !sf )
		{
			BB_log( BB_LOG_WARNING, "[audiosrcfile]: cannot open file '%s'...", filename );
			m_lasterror = "cannot open file!";
			return FALSE;
		}

		// support mono and stereo only
		if( sf_info.channels > 2 )
		{
		    BB_log( BB_LOG_SEVERE, "[audiosrcfile]: not mono/stereo file!" );
		    m_lasterror = "not a mono/stereo file";
			this->close();
			return FALSE;
		}

		if( sf_info.samplerate != BirdBrain::srate() )
		{
#ifdef SECRET_RABBIT_CODE
			sr_gain = 1.0f;
			int converted;
			do {
				converted = samplerateconvert( filename );
			}
			while( converted < 0 );
#else
			BB_log( BB_LOG_SEVERE, "[audiosrcfile]: not %i hz file!", BirdBrain::srate() );
			m_lasterror = "TAPESTREA expects matching sample rate (";
			char buffy[128];
			sprintf( buffy, "%d", BirdBrain::srate() );
			m_lasterror += buffy;
			m_lasterror += "), for now...";
			this->close();
			return FALSE;
#endif
		}
		
		// add to cachemanager
		if( usecache )
			AudioCentral::instance()->m_cachemanager->addsf( filename, SFM_READ, sf_info, sf, id ); // mode could be wrong but needs work anyway
	}

    // check start and len
    if( start >= sf_info.frames )
    {
        BB_log( BB_LOG_SEVERE, "[audiosrcfile]: '%s' has only %li frames...",
            filename, sf_info.frames );
        BB_log( BB_LOG_SEVERE, "[audiosrcfile]: ...cannot seek to '%li'...",
            start );
        m_lasterror = "cannot seek to requested start position";
        this->close();
        return FALSE;
    }

    if( len && (start + len) > sf_info.frames ) // used to be >= but... see m_end below
    {
        BB_log( BB_LOG_SEVERE, "[audiosrcfile]: '%s' has only %li frames...",
            filename, sf_info.frames );
        BB_log( BB_LOG_SEVERE, "[audiosrcfile]: ...cannot seek upto '%li'...",
            start + len );
        m_lasterror = "cannot seek to requested start position";
        this->close();
        return FALSE;
    }

    // copy
    m_filename = filename;
    m_start = start;
    m_end = len ? start + len : sf_info.frames;

    if( scale )
    {
        // get scaling
        double   max_val ;
        sf_command (sf, SFC_CALC_NORM_SIGNAL_MAX, &max_val, sizeof (max_val)) ;
        signal_scale = (float)( .95 / max_val );
    }

    // seek
    if( start >= 0 ) sf_seek( sf, start, SEEK_SET );
    
    return TRUE;
}



#define SR_BUFFER_LEN 4096
#ifdef SECRET_RABBIT_CODE
//-----------------------------------------------------------------------------
// name: samplerateconvert()
// desc: duh
//-----------------------------------------------------------------------------
int AudioSrcFile::samplerateconvert( const char * filename )
{
	SNDFILE * outfile; 
	SF_INFO outfileinfo = sf_info;
	outfileinfo.samplerate = BirdBrain::srate();
	outfileinfo.sections = 0;
	outfileinfo.frames = 0;
	outfileinfo.format = SF_FORMAT_WAV | SF_FORMAT_FLOAT; 
	int error;

	BB_log( BB_LOG_INFO, "Converting sample rate of '%s' from %i Hz to %i Hz", 
		filename, (int)sf_info.samplerate, (int)outfileinfo.samplerate );
	// open converter
	if( sr )
	{
		BB_log( BB_LOG_WARNING, "[audiosrcfile]: surprise, sr already existed, quitting" );
		return 0;
	}
	sr = src_new( SRC_SINC_FASTEST, outfileinfo.channels, &error );
	if( !sr )
	{
		BB_log( BB_LOG_WARNING, "[audiosrcfile]: %s", src_strerror( error ) );
		return 0;
	}

	// open outfile
	/*char temp[8];
	std::string outfilename = BirdBrain::getname( filename ) + "_temp_" + itoa((int)BirdBrain::srate(),temp,10) + ".wav"; 
	if ( (outfile = sf_open( outfilename.c_str() , SFM_WRITE, &outfileinfo )) == NULL)
	{	
		BB_log( BB_LOG_SEVERE, "[audiosrcfile]: Not able to open output file '%s'; not converting", outfilename.c_str() );
		return 0;
	}*/
	FILE * temp;
	if( (temp = tmpfile()) == NULL )
	{
		BB_log( BB_LOG_SEVERE, "[audiosrcfile]: Could not open temp file; not converting" );
		src_delete( sr ); sr = NULL;
		return 0;
	}
	int temporary;
#ifdef __PLATFORM_WIN32__
	temporary = _fileno( temp );
#else
	temporary = fileno( temp );
#endif
	outfile = sf_open_fd( temporary, SFM_RDWR, &outfileinfo, TRUE );
	if( outfile == NULL )
	{
		BB_log( BB_LOG_SEVERE, "[audiosrcfile]: Error opening sound file from descriptor %i (%s); not converting", 
			temporary, sf_error_number( sf_error( sf ) ) );
		fclose( temp );
		src_delete( sr ); sr = NULL;
		return 0;
	}
	else
	{
#ifdef __PLATFORM_WIN32__
		BB_log( BB_LOG_INFO, "[audiosrcfile]: Opened temporary file %i", _fileno( temp ) );
#else
		BB_log( BB_LOG_INFO, "[audiosrcfile]: Opened temporary file %i", fileno( temp ) );
#endif
	}

	// convert (taken from libsamplerate example sndfile-resample.c)
	SAMPLE input [SR_BUFFER_LEN] ;
	SAMPLE output [SR_BUFFER_LEN] ;
	SRC_DATA src_data ;
	sf_count_t output_count = 0 ;

	sf_seek (sf, 0, SEEK_SET);
	sf_seek (outfile, 0, SEEK_SET);
	src_data.end_of_input = 0; // Set this later

	// Start with zero to force load in while loop. 
	src_data.input_frames = 0;
	src_data.data_in = input;
	src_data.src_ratio = (double)outfileinfo.samplerate / sf_info.samplerate; 
	src_data.data_out = output ;
	src_data.output_frames = SR_BUFFER_LEN / sf_info.channels ;

	while( true )
	{
		// If the input buffer is empty, refill it. 
		if (src_data.input_frames == 0)
		{	src_data.input_frames = sf_readf_float (sf, input, SR_BUFFER_LEN / sf_info.channels);
			src_data.data_in = input;

			// The last read will not be a full buffer, so end_of_input.
			if (src_data.input_frames < SR_BUFFER_LEN / sf_info.channels)
				src_data.end_of_input = SF_TRUE;
		}

		if ( (error = src_process( sr, &src_data )) )
		{	
			BB_log( BB_LOG_SEVERE, "[audiosrcfile]: %s - not converting", src_strerror( error ) );
			sf_close( outfile );
			return 0;
		}

		// Terminate if done.
		if (src_data.end_of_input && src_data.output_frames_gen == 0)
			break;
		
		// apply gain if needed
		float max = 1;
		for( int k = 0; k < src_data.output_frames_gen * sf_info.channels; k++ )
		{
			output[k] *= sr_gain;
			if( ::fabs( output[k] ) > max )
				max = ::fabs( output[k] );
		}
		if( max > 1 )
		{
			sr_gain *= 1 / max;
			BB_log( BB_LOG_WARNING, "[audiosrcfile]: output clipped (max = %f, gain = %f), redoing conversion", max, sr_gain );
			sf_close( outfile );
			src_delete( sr ); sr = NULL;
			return -1;
		}

		// Write output
		sf_writef_float( outfile, output, src_data.output_frames_gen );
		output_count += src_data.output_frames_gen;

		// Update read info
		src_data.data_in += src_data.input_frames_used * sf_info.channels;
		src_data.input_frames -= src_data.input_frames_used;
	}

//	sr = src_delete (sr);

	// close things
	//sf_close( outfile );
	sf_close( sf );
	src_delete( sr ); sr = NULL;

	sf_command( outfile, SFC_UPDATE_HEADER_NOW, NULL, 0 );
	sf_info = outfileinfo;
	sf_info.frames = sf_seek( outfile, 0, SEEK_CUR );
	BB_log( BB_LOG_CRAZY, "sf frames : %i", (int)sf_info.frames );
	BB_log( BB_LOG_CRAZY, "outfile seekable %i sf_info seekable %i", (int)outfileinfo.seekable, (int)sf_info.seekable );
	sf = outfile;

	// change sf to be the file we just wrote
	//sf_info = outfileinfo;
	/*if ( !(sf = sf_open( outfilename.c_str(), SFM_READ, &sf_info )) )
	{	
		BB_log( BB_LOG_SEVERE, "[audiosrcfile]: Not able to open temporary input file '%s'", outfilename.c_str() );
		return 0;
	} */
	/*if( (sf = sf_open_fd( _fileno( temp ), SFM_READ, &sf_info, TRUE )) == NULL )
	{
		BB_log( BB_LOG_SEVERE, "[audiosrcfile]: Not able to open temporary input file %i: %s", 
			_fileno( temp ), sf_error_number( sf_error( sf ) ) );
		fclose( temp );
		return 0;
	}*/
	
	return (int)output_count;
}
#endif


//-----------------------------------------------------------------------------
// name: close()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioSrcFile::close()
{
    if( !sf ) return FALSE;

    // deallocate buffer
    if( m_buffer ) delete [] m_buffer;
    // close
//    sf_close( sf );
	AudioCentral::instance()->m_cachemanager->closesf( sf );
	// delete s.r.c

#ifdef SECRET_RABBIT_CODE
	if( sr ) 
	{
		src_delete( sr );
		sr = NULL;
	}
#endif

    m_buffer = NULL;
    sf = NULL;
    m_start = 0;
    m_end = 0;
    m_filename = "";
    m_buffer_size = 0;

    return TRUE;
}




//-----------------------------------------------------------------------------
// name: stick()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioSrcFile::stick( SAMPLE * buffer, t_TAPUINT num_frames )
{
//    assert( m_done == FALSE );

    // are we done?
    if( m_stop_asap )
        return FALSE;

    // do mono first
    t_TAPBOOL it = mtick( buffer, num_frames );
    // convert to "stereo"
    mono2stereo( buffer, num_frames, m_buffer );
    // pan
    pan_buf( buffer, num_frames, m_pan );
    // copy back
    memcpy( buffer, m_buffer, sizeof(SAMPLE) * num_frames * 2 );

    return it;
}


//-----------------------------------------------------------------------------
// name: mtick()
// desc: mono tick
//-----------------------------------------------------------------------------
t_TAPBOOL AudioSrcFile::mtick( SAMPLE * buffer, t_TAPUINT num_frames )
{
//    assert( m_done == FALSE );

    // are we done?
    if( m_stop_asap )
        return FALSE;

	// zero
	memset( buffer, 0, num_frames * sizeof(SAMPLE) );

    // sanity
    if( !sf )
    {
        return FALSE;
    }

    sf_count_t i;
    if( m_buffer_size < ( num_frames * 2 ) )
    {
        // delete
        if( m_buffer ) delete [] m_buffer;
        // allocate
        m_buffer = new SAMPLE[num_frames * 2];
        if( !m_buffer )
        {
            fprintf( stderr, "[audiosrcfile]: cannot allocate buffer of size '%i'...\n",
                num_frames * 2 );
            return FALSE;
        }
        m_buffer_size = num_frames * 2;
    }

    // read the next chunk
    sf_count_t diff = m_end - sf_seek( sf, 0, SEEK_CUR );
    // check to make sure there is enough data from this point
    if( diff <= 0 )
    {
        if( slide_locally ) *slide_locally = false;
        return FALSE;
    }

	// hack
	sf_count_t pos = sf_seek( sf, 0, SEEK_CUR );
    // read
    sf_count_t thismuch = num_frames <= diff ? num_frames : diff;
    sf_count_t read = sf_readf_float( sf, m_buffer, thismuch );
    //assert( thismuch == read );
	if( thismuch != read )
	{
		BB_log( BB_LOG_FINE, "(audiosrcfile): couldn't read enough; %s", sf_error_number(sf_error(sf)) );
	}
	BB_log( BB_LOG_FINEST, "To read: %i; Read: %i", (int)thismuch, (int)read );
	// hack continued, possibly the correct action
	if( sf_seek( sf, 0, SEEK_CUR ) == pos )
	{
		BB_log( BB_LOG_FINEST, "sf_read no update cur pointer; me do manually :|" );
		sf_seek( sf, read, SEEK_CUR );
	}

    // update
    if( slide != NULL )
    {
        *slide = (sf_seek( sf, 0, SEEK_CUR ) - m_start) / (t_TAPSINGLE)(m_end - m_start);
		BB_log( BB_LOG_FINE, "slide = (%i - %i) / (%i - %i) = %f",
			(int)(sf_seek( sf, 0, SEEK_CUR )), (int)m_start, (int)m_end, (int)m_start, *slide );
        *slide_locally = true;
    }

    // if more than one channel
    if( sf_info.channels > 1 )
    {
        // convert multi channel to mono
        stereo2mono( m_buffer, num_frames, buffer );
    }
    else
    {
        // copy
        memcpy( buffer, m_buffer, sizeof(SAMPLE) * thismuch );
    }


    // scale
    for( i = 0; i < thismuch; i++ )
        buffer[i] *= signal_scale;

    // gain
    gain_buf( buffer, num_frames, m_gain );

    return TRUE;
}





//-----------------------------------------------------------------------------
// name: stick()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioSrcFrame::stick( SAMPLE * buffer, t_TAPUINT num_frames )
{
//    assert( m_done == FALSE );

    // are we done?
    if( m_stop_asap )
        return FALSE;

    // more data?
    if( pos >= frame->wsize )
    {
        if( slide_locally ) *slide_locally = false;
        return FALSE;
    }

    // copy
    t_TAPUINT thismuch = (frame->wsize - pos) > num_frames ? num_frames : (frame->wsize - pos);
    SAMPLE * src = frame->waveform + pos;
    SAMPLE * end = src + thismuch;
    SAMPLE * dest = buffer;
    while( src < end )
    {
        *dest = *(dest+1) = *src;
        src++;
        dest += 2;
    }

    // zero 
    if( thismuch < num_frames )
        for( t_TAPUINT i = thismuch; i < num_frames; i++ )
        {
            buffer[2*i] = 0.0f;
            buffer[2*i+1] = 0.0f;
        }

    // advance pos
    pos += thismuch;

    // update
    if( slide != NULL )
    {
        *slide = pos / (t_TAPSINGLE)frame->wsize;
        *slide_locally = true;
    }

    // gain
    gain_buf( buffer, num_frames * 2, m_gain );
    // pan
    pan_buf( buffer, num_frames, m_pan );

    return TRUE;
}



//-----------------------------------------------------------------------------
// name: mtick()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioSrcFrame::mtick( SAMPLE * buffer, t_TAPUINT num_samples )
{
//    assert( m_done == FALSE );

    // are we done?
    if( m_stop_asap )
        return FALSE;

    // more data?
    if( pos >= frame->wsize )
    {
        if( slide_locally ) *slide_locally = false;
        return FALSE;
    }

    // copy
    t_TAPUINT thismuch = (frame->wsize - pos) > num_samples ? num_samples : (frame->wsize - pos);
    SAMPLE * src = frame->waveform + pos;
    SAMPLE * dest = buffer;
    
    memcpy( dest, src, thismuch * sizeof(SAMPLE) ); 

    // zero 
    if( thismuch < num_samples )
        memset( buffer + thismuch, 0, (num_samples - thismuch) * sizeof(SAMPLE) ); 

    // advance pos
    pos += thismuch;

    // update
    if( slide != NULL )
    {
        *slide = pos / (t_TAPSINGLE)frame->wsize;
        *slide_locally = true;
    }

    // gain
    gain_buf( buffer, num_samples, m_gain );

    return TRUE;
}


//-----------------------------------------------------------------------------
// name: seek()
// desc: audiosrcframe seek
//-----------------------------------------------------------------------------
t_TAPBOOL AudioSrcFrame::seek( t_TAPUINT where, t_TAPUINT how )
{
    // checks
    if( how == SEEK_CUR )
        where += pos; 
    else if( how != SEEK_SET )
        return FALSE;
    if( where < 0 || where >= frame->wsize )
        return FALSE;
    if( this->playing() )
        return FALSE;

    // gist
    pos = where;
    return TRUE;
}


// ...
int AudioSrcFrame::num_samples()
{
    if( frame )
        return frame->wsize;
    else
        return 0;
}



//-----------------------------------------------------------------------------
// name: stick()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioSrcEliot::stick( SAMPLE * buffer, t_TAPUINT num_frames )
{
//    assert( m_done == FALSE );

    // are we done?
    if( m_stop_asap )
        return FALSE;

    memset( buffer, 0, sizeof(SAMPLE) * num_frames );
    ts_io->m_audio_cb( (char *)buffer, num_frames, NULL );
    // convert to "stereo" in-place
    mono2stereo( buffer, num_frames );
    // gain
    gain_buf( buffer, num_frames * 2, m_gain );
    // pan
    pan_buf( buffer, num_frames, m_pan );


    return TRUE;
}




//-----------------------------------------------------------------------------
// name: set()
// desc: ...
//-----------------------------------------------------------------------------
void AudioFxReverb::set( t_TAPUINT which, t_TAPFLOAT T60 )
{
    // do nothing
    if( which == m_state ) return;

    if( m_left ) SAFE_DELETE( m_left );
    if( m_right ) SAFE_DELETE( m_right );

    // no reverb
    if( which == NONE )
    { } // do nothing
    else if( which == PRC )
    {
        m_left = new FxPRCRev( T60 );
        m_right = new FxPRCRev( T60 );
    }
    else if( which == JC )
    {
        m_left = new FxJCRev( T60 );
        m_right = new FxJCRev( T60 );
    }
    else if( which == N )
    {
        m_left = new FxNRev( T60 );
        m_right = new FxNRev( T60 );
    }
    else
        // set to none
        set( NONE, 0 );

    // set mix
    mix( m_mix );
}




//-----------------------------------------------------------------------------
// name: mix()
// desc: ...
//-----------------------------------------------------------------------------
void AudioFxReverb::mix( t_TAPFLOAT value )
{
    m_mix = value;
    if( m_left ) m_left->setEffectMix( value );
    if( m_right ) m_right->setEffectMix( value );
}




//-----------------------------------------------------------------------------
// name: stick()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioFxReverb::stick( SAMPLE * buffer, t_TAPUINT num_frames )
{
    if( m_left && m_mix > 0.001 )
    {
        // do left
        for( t_TAPUINT i = 0; i < num_frames; i++ )
        {
            buffer[i*2] = m_left->tick( buffer[i*2] );
        }
    }

    if( m_right && m_mix > 0.001 )
    {
        // do right
        for( t_TAPUINT i = 0; i < num_frames; i++ )
        {
            buffer[i*2+1] = m_right->tick( buffer[i*2+1] );
        }
    }

    return TRUE;
}




//-----------------------------------------------------------------------------
// name: init()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPUINT AudioBus::init( t_TAPUINT num_frames )
{
    m_num_frames = num_frames;
	m_max_channels = 9; 
    m_buffer = new SAMPLE[num_frames*2];
    m_buffer2 = new SAMPLE[num_frames*2];
	m_bufferM = new SAMPLE[num_frames * m_max_channels]; 
	memset( m_buffer, 0, sizeof(SAMPLE) * num_frames * 2 );
    m_ready = FALSE;
    m_done = FALSE;
    m_gain = 1.0f;
    on = NULL;

    // allocate reverb
    m_reverb.set( AudioFxReverb::N, 4.0 );

    return TRUE;
}



//-----------------------------------------------------------------------------
// name: ~AudioBus()
// desc: ...
//-----------------------------------------------------------------------------
AudioBus::~AudioBus()
{
    SAFE_DELETE_ARRAY( m_buffer );
    SAFE_DELETE_ARRAY( m_buffer2 );
	SAFE_DELETE_ARRAY( m_bufferM );
}




//-----------------------------------------------------------------------------
// name: stick()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioBus::stick( SAMPLE * buffer, t_TAPUINT num_frames )
{
//    assert( m_done == FALSE );

    // buses always return true
    if( m_src.size() == 0 )
    {
        memset( buffer, 0, sizeof(SAMPLE) * num_frames * 2 );
        // need to do reverb
        m_reverb.stick( buffer, num_frames );
        return TRUE;
    }

    SAMPLE * buf = buffer ? buffer : m_buffer;
    SAMPLE * buf2 = m_buffer2;
    t_TAPUINT size = num_frames ? num_frames : m_num_frames;
    t_TAPBOOL yes = FALSE;

    // zero out the buffer
    memset( buf, 0, sizeof(SAMPLE) * size * 2 );
    
    assert( num_frames == 0 || size <= m_num_frames );

    // lock
    m_mutex.acquire();
    // clear
    m_src2.clear();
	// loop
    for( int i = 0; i < m_src.size(); i++ )
    {
        // get samples
        yes = m_src[i]->stick( buf2, size );
        if( !yes )
        {
            // log
            BB_log( BB_LOG_FINE, "audio source '%s' done...", m_src[i]->m_name.c_str() );

            if( m_src[i]->m_delete )
                delete m_src[i];
            else
                m_src[i]->m_done = TRUE;
        }
        else
        {
            // if there is more
            m_src2.push_back( m_src[i] );
            // add
            for( t_TAPUINT j = 0; j < size*2; j++ )
                buf[j] += buf2[j];
        }
    }
    // gain
    gain_buf( buf, size*2, m_gain );
    // pan
    pan_buf( buf, size, m_pan );
    // reverb
    m_reverb.stick( buf, size );
    // copy
    if( buffer && buffer != buf )
        memcpy( buffer, buf, sizeof(SAMPLE) * num_frames * 2 );
    // more src
    m_src = m_src2;
    // release
    m_mutex.release();

    // no return false because this is a bus
    return TRUE;
}


//-----------------------------------------------------------------------------
// name: multitick()
// desc: stick into stereobuffer, multiple single channels into multibuffer
//-----------------------------------------------------------------------------
t_TAPBOOL AudioBus::multitick( SAMPLE * stereobuffer, SAMPLE * multibuffer, t_TAPUINT num_frames )
{
//    assert( m_done == FALSE );

    // buses always return true
    if( m_src.size() == 0 )
    {
        memset( stereobuffer, 0, sizeof(SAMPLE) * num_frames * 2 );
        // memset( multibuffer, 0, sizeof(SAMPLE) * num_frames * m_src.size() );
		// need to do reverb
        m_reverb.stick( stereobuffer, num_frames );
        return TRUE;
    }

    SAMPLE * buf = stereobuffer ? stereobuffer : m_buffer;
    SAMPLE * buf2 = m_buffer2;
	SAMPLE * bufM = multibuffer; 
	int M = m_src.size(); 
    t_TAPUINT size = num_frames ? num_frames : m_num_frames;
    t_TAPBOOL yes = FALSE;

    // zero out the buffers
    memset( buf, 0, sizeof(SAMPLE) * size * 2 );
	if( bufM ) memset( bufM, 0, sizeof(SAMPLE) * size * M );


    assert( num_frames == 0 || size <= m_num_frames );

    // lock
    m_mutex.acquire(); 
	// clear
    m_src2.clear();
    // loop
    for( int i = 0; i < M; i++ )
    {
        // get samples
        yes = m_src[i]->stick( buf2, size );
        if( !yes )
        {
            // log
            BB_log( BB_LOG_FINE, "audio source '%s' done...", m_src[i]->m_name.c_str() );

            if( m_src[i]->m_delete )
                delete m_src[i];
            else
                m_src[i]->m_done = TRUE;
        }
        else
        {
            // if there is more
            m_src2.push_back( m_src[i] );
            // add for stereo buffer
            for( t_TAPUINT j = 0; j < size*2; j++ )
                buf[j] += buf2[j];
			// pre-process + interleave for multibuffer
			if( bufM )
			{
				gain_buf( buf2, size*2, m_gain ); 
				pan_buf( buf2, size, m_pan );
				//m_reverb.stick( buf2, size );
				int source = i;
				for( t_TAPUINT sample = 0; sample < size; sample++ )
					bufM[sample*M + source] = (buf2[2*sample] + buf2[2*sample+1])/2;
			}
        }
    }
	// post-process for stereo buffer
    // gain
    gain_buf( buf, size*2, m_gain );
    // pan
    pan_buf( buf, size, m_pan );
    // reverb
    m_reverb.stick( buf, size );
    // copy
    if( stereobuffer && stereobuffer != buf )
        memcpy( stereobuffer, buf, sizeof(SAMPLE) * num_frames * 2 );
    // more src
    m_src = m_src2;
    // release
    m_mutex.release();

    // no return false because this is a bus
    return TRUE;
}


//-----------------------------------------------------------------------------
// name: play()
// desc: ...
//-----------------------------------------------------------------------------
void AudioBus::play( AudioSrc * src, t_TAPBOOL solo, t_TAPBOOL rew )
{
    // solo
    if( solo ) this->stop();

    // synchronize
    m_mutex.acquire();
    src->reset();
    if( rew )
        // rewind it
        src->rewind();
    //else
        // reset it
        //src->reset(); // redundant
    // add to vector
    t_TAPBOOL got = FALSE;
    for( t_TAPINT i = 0; i < m_src.size(); i++ )
        if( m_src[i] == src ) { got = TRUE; break; }
    if( !got )
	{
        m_src.push_back( src );
//		fprintf( stdout, "Bus %x push back %x\n", this, src );
//		for( t_TAPINT i = 0; i < m_src.size(); i++ )
//			fprintf( stdout, "  %x\n", m_src[i] );
	}
    // release
    m_mutex.release();
}




//-----------------------------------------------------------------------------
// name: play()
// desc: ...
//-----------------------------------------------------------------------------
void AudioBus::remove( AudioSrc * src )
{
    // synchronize
    m_mutex.acquire();
//	vector<AudioSrc *> temp;
    // clear src2
    m_src2.clear();
    // copy
    for( t_TAPUINT i = 0; i < m_src.size(); i++ )
    {
        // match
        if( m_src[i] != src )
            m_src2.push_back( m_src[i] );
//        if( m_src[i] != src )
//            temp.push_back( m_src[i] );
//		if( m_src[i] == src )
//			m_src.erase( m_src.begin() + i-- ); 
    }
    // copy
	m_src = m_src2;
	m_src2.clear();
	//m_src = temp;
    // synchronize
    m_mutex.release();
}




//-----------------------------------------------------------------------------
// name: stop()
// desc: ...
//-----------------------------------------------------------------------------
void AudioBus::stop()
{
    // synchronize
    m_mutex.acquire();
    // delete
    for( long i = 0; i < m_src.size(); i++ )
    {
        // log
        BB_log( BB_LOG_FINE, "audio source '%s' done...", m_src[i]->m_name.c_str() );

        if( m_src[i]->m_delete )
            delete m_src[i];
        else
            m_src[i]->m_done = TRUE;
    }

    // clear
    m_src.clear();
    // more
    m_ready = FALSE;
    // release
    m_mutex.release();
}




//-----------------------------------------------------------------------------
// name: the_parallel()
// desc: (METHOD A)
//-----------------------------------------------------------------------------
#ifdef __PLATFORM_WIN32__
unsigned __stdcall the_parallel( void * data )
#else
void * the_parallel( void * data )
#endif
{
    AudioBusParallel * parallel = (AudioBusParallel *)data;

    // loop until done
    while( parallel->m_done == FALSE )
    {
        // wait for signal
        while( parallel->m_ready )
            usleep( 1000 );

        // accumulate
        parallel->accumulate();

        // wait
        parallel->m_ready = TRUE;
    }

    return 0;
}




//-----------------------------------------------------------------------------
// name: AudioBusParallel()
// desc: ...
//-----------------------------------------------------------------------------
AudioBusParallel::AudioBusParallel( t_TAPUINT num_frames, t_TAPUINT sig_frames )
    : AudioBus( num_frames )
{
    sf_stereo = NULL;
	sf_multi = NULL;
    m_sig_frames = sig_frames;
    // METHOD A: start thread
    // METHOD A: m_thread.start( the_parallel, this );
}


bool AudioBusParallel::open( int mode, const string & filename, const string & filename2 )
{
	bool ret;

	if( mode == AudioCentral::STEREO || mode == AudioCentral::BOTH )
	{
		m_sf_stereo_mutex.acquire();
		memset( &sf_stereo_info, 0, sizeof(sf_stereo_info) );
		sf_stereo_info.samplerate = BirdBrain::srate();
		sf_stereo_info.channels = 2;
		sf_stereo_info.format = SF_FORMAT_WAV | SF_FORMAT_PCM_16;
		if( sf_stereo )
			AudioCentral::instance()->m_cachemanager->closesf( sf_stereo );
		sf_stereo = AudioCentral::instance()->m_cachemanager->opensf( filename.c_str(), SFM_WRITE, &sf_stereo_info );
		if( sf_stereo )
			BB_log( BB_LOG_INFO, "AudioBusParallel: opened write file, %i channels", sf_stereo_info.channels );
		else
			BB_log( BB_LOG_INFO, "AudioBusParallel: failed open, %i channels", sf_stereo_info.channels );
		ret = (sf_stereo != NULL);
		m_sf_stereo_mutex.release();
	}
	if( (mode == AudioCentral::MULTICHANNEL || mode == AudioCentral::BOTH) && m_src.size() > 0 )
	{
		m_sf_multi_mutex.acquire();
		memset( &sf_multi_info, 0, sizeof(sf_multi_info) );
		sf_multi_info.samplerate = BirdBrain::srate();
		sf_multi_info.channels = m_src.size();
		sf_multi_info.format = SF_FORMAT_WAV | SF_FORMAT_PCM_16;
		if( sf_multi )
			AudioCentral::instance()->m_cachemanager->closesf( sf_multi );
		sf_multi = AudioCentral::instance()->m_cachemanager->opensf( 
			( mode == AudioCentral::BOTH ) ? filename2.c_str() : filename.c_str(), 
			SFM_WRITE, &sf_multi_info );
		if( sf_multi )
			BB_log( BB_LOG_INFO, "AudioBusParallel: opened write file, %i channels", sf_multi_info.channels );
		else
			BB_log( BB_LOG_INFO, "AudioBusParallel: failed open, %i channels", sf_multi_info.channels );
		ret = ret && (sf_multi != NULL);
		m_sf_multi_mutex.release();
	}

//    sf = sf_open( filename.c_str(), SFM_WRITE, &sf_info );

    return ret;
}


// close all
void AudioBusParallel::close()
{
	m_sf_stereo_mutex.acquire(); 
	if( sf_stereo )
	{
		AudioCentral::instance()->m_cachemanager->closesf( sf_stereo );
		sf_stereo = NULL;
	}
	m_sf_stereo_mutex.release();
	m_sf_multi_mutex.acquire();
	if( sf_multi )
	{
		AudioCentral::instance()->m_cachemanager->closesf( sf_multi );
		sf_multi = NULL;
	}
	m_sf_multi_mutex.release(); 
//    if( sf )
////        sf_close( sf );
//		AudioCentral::instance()->m_cachemanager->closesf( sf );
//    sf = NULL;
}

void AudioBusParallel::close( int mode )
{
    m_sf_stereo_mutex.acquire();
	if( (mode == AudioCentral::STEREO || mode == AudioCentral::BOTH) && sf_stereo )
	{
		AudioCentral::instance()->m_cachemanager->closesf( sf_stereo );
		sf_stereo = NULL;
	}
	m_sf_stereo_mutex.release();
	m_sf_multi_mutex.acquire();
	if( (mode == AudioCentral::MULTICHANNEL || mode == AudioCentral::BOTH) && sf_multi )
	{
		AudioCentral::instance()->m_cachemanager->closesf( sf_multi );
		sf_multi = NULL;
	}
	m_sf_multi_mutex.release();
}




//-----------------------------------------------------------------------------
// name: stick()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioBusParallel::stick( SAMPLE * buffer, t_TAPUINT num_frames )
{
//    assert( m_done == FALSE );

    if( m_src.size() == 0 )
    {
        memset( buffer, 0, sizeof(SAMPLE) * num_frames * 2 );
        return TRUE;
    }

    // METHOD A: wait for data
    // while( !m_ready )
    //     usleep( 2000 );

    // METHOD B: callback (no blocking)
    accumulate();

    // copy
    memcpy( buffer, m_buffer, sizeof(SAMPLE) * num_frames * 2 );

    // METHOD A: signal (no harm to keep in METHOD B)
    m_ready = FALSE;

    return TRUE;
}


//-----------------------------------------------------------------------------
// name: multitick()
// desc: ...
//-----------------------------------------------------------------------------
t_TAPBOOL AudioBusParallel::multitick( SAMPLE * stereobuffer, SAMPLE * multibuffer, t_TAPUINT num_frames )
{
//    assert( m_done == FALSE );

    if( m_src.size() == 0 )
    {
        memset( stereobuffer, 0, sizeof(SAMPLE) * num_frames * 2 );
        return TRUE;
    }

    // METHOD A: wait for data
    // while( !m_ready )
    //     usleep( 2000 );

    // METHOD B: callback (no blocking)
    accumulate();

    // copy
    memcpy( stereobuffer, m_buffer, sizeof(SAMPLE) * num_frames * 2 );
	memcpy( multibuffer, m_bufferM, sizeof(SAMPLE) * num_frames * m_src.size() );

    // METHOD A: signal (no harm to keep in METHOD B)
    m_ready = FALSE;

    return TRUE;
}



//-----------------------------------------------------------------------------
// name: accumulate()
// desc: ...
//-----------------------------------------------------------------------------
void AudioBusParallel::accumulate( )
{
    t_TAPUINT togo = m_num_frames;
    t_TAPUINT num_frames = m_sig_frames;

    // loop
    while( togo > 0 )
    {
        // check
        if( num_frames > togo ) num_frames = togo;

        // tick sig
        //AudioBus::stick( m_buffer + (m_num_frames-togo)*2, num_frames );
    	int channels = m_src.size() < m_max_channels ? m_src.size() : m_max_channels; 
		AudioBus::multitick( m_buffer + (m_num_frames-togo)*2, 
							 m_bufferM + (m_num_frames-togo) * channels,  
							 num_frames );

        // keep track of frames to go
        togo -= num_frames;
    }

    // for performance reasons (we are not stupid)
    if( sf_stereo )
    {
        m_sf_stereo_mutex.acquire();
        // in METHOD B, this may cause more audio clicks (maybe)
        if( sf_stereo ) sf_writef_float( sf_stereo, m_buffer, m_num_frames );
        m_sf_stereo_mutex.release();
    }
	if( sf_multi )
    {
        m_sf_multi_mutex.acquire();
        // in METHOD B, this may cause more audio clicks (maybe)
        if( sf_multi ) sf_writef_float( sf_multi, m_bufferM, m_num_frames );
        m_sf_multi_mutex.release();
    }
}


// what's going on???
SNDFILE * CacheManager::opensf( std::string path, int mode, SF_INFO * info, int id, t_TAPBOOL usecache )
{
	// allowed to use cached version, and it's already open
	if( usecache && mode != SFM_WRITE && isopen( path ) )
	{
		BB_log( BB_LOG_FINE, "[CacheManager]: '%s' was already open", path.c_str() );
		std::vector<CacheInfo> *civ = &(names[path]);
		if( id < 0 )
			id = 0;
		if( id < civ->size() )
		{
			CacheInfo ci = civ->at( id );
			ci.count++;
			ci.close = false;
			*info = ci.lastinfo;
			// rewind
			sf_seek( ci.sf, 0, SEEK_SET );
			(*civ)[id] = ci;
			return ci.sf; 
		}
	}
	// need to open
	SNDFILE * sf = sf_open( path.c_str(), mode, info );
	if( sf == NULL )
	{
		BB_log( BB_LOG_WARNING, "[CacheManager]: could not open '%s'; %s", path.c_str(), 
			sf_error_number( sf_error( sf ) ) );
	}
	else
	{
		if( mode == SFM_WRITE || !usecache )
			return sf;
		// otherwise
		CacheInfo ci; 
		ci.sf = sf;
		ci.count = 1;
		ci.close = false;
		names[path].push_back( ci );
		BB_log( BB_LOG_INFO, "[CacheManager]: opening '%s'", path.c_str() );
		// rewind
		sf_seek( sf, 0, SEEK_SET );
	}
	return sf;
}


// must not already be open with this id
bool CacheManager::addsf( std::string path, int mode, const SF_INFO &info, SNDFILE * sf, int id )
{
	if( isopen( path ) > id )
	{
		BB_log( BB_LOG_WARNING, "[CacheManager]: '%s' already open, not adding", path.c_str() );
		return false;
	}
	else if( mode == SFM_WRITE )
	{
		BB_log( BB_LOG_WARNING, "[CacheManager]: write only file not included in cache" );
		return false;
	}
	else
	{
		CacheInfo ci; 
		ci.sf = sf;
		ci.close = false;
		ci.count = 1; 
		ci.lastinfo = info;
		std::vector<CacheInfo> * civ = &(names[path]);
		if( id < civ->size() )
			(*civ)[id] = ci;
		else
		{
			civ->push_back( ci );
		}
		BB_log( BB_LOG_INFO, "[CacheManager]: adding '%s'", path.c_str() );
	}
	return true;
}


void CacheManager::closesf( SNDFILE * sf )
{
	std::map<std::string, std::vector<CacheInfo> >::iterator iter;
	bool found = false;
	for( iter = names.begin(); iter != names.end(); )
	{
		std::vector<CacheInfo> * civ = &(iter->second);
		bool erased = false;
		for( int i = 0; i < civ->size(); i++ )
		{
			CacheInfo ci = (*civ)[i];
			if( ci.sf == sf )
			{
				found = true;
				std::string path = iter->first;
				if( --ci.count < 0 ) ci.count = 0;
				if( ci.close && ci.count == 0 )
				{
					BB_log( BB_LOG_INFO, "[CacheManager]: closing '%s'", path.c_str() );
					sf_close( sf ); 
					// erase
					civ->erase( civ->begin() + i-- );
					if( civ->empty() )
					{
						names.erase( iter++ );
						erased = true;
					}
				}
				else
				{
					// update
					(*civ)[i] = ci;
					BB_log( BB_LOG_FINE, "[CacheManager]: '%s' not closed just in case it will be used again", path.c_str() );
				}
				break;
			}
		}
		if( !erased ) ++iter;
	}
	// may not be cached, still needs to be closed
	if( !found )
		sf_close( sf );
}


int CacheManager::isopen( std::string path )
{
	std::map<std::string, std::vector<CacheInfo> >::iterator iter = names.find( path );
	if( iter != names.end() )
		return iter->second.size();
	else
		return 0;
}


// AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH!!!!!!!!!!!!!!!!!!!
